# Tests for "symbex --async / --class / --typed etc"
import pathlib
import pytest
from click.testing import CliRunner

from symbex.cli import cli


@pytest.mark.parametrize(
    "args,expected",
    (
        (
            ["--function"],
            [
                "def func_no_args",
                "def func_positional_args",
                "async def async_func",
                "def func_default_args",
                "def func_arbitrary_positional_args",
                "def func_arbitrary_keyword_args",
                "def func_arbitrary_args",
                "def func_positional_only_args",
                "def func_keyword_only_args",
                "def func_type_annotations",
                "def function_with_non_pep_0484_annotation",
                "def complex_annotations",
                "def func_fully_typed",
                "async def async_func_fully_typed",
                "def func_partially_typed",
                "def func_partially_typed_no_typed_return",
                "def func_partially_typed_only_typed_return",
                "def func_typed_no_params",
            ],
        ),
        (
            ["--class"],
            [
                "class ClassNoBase",
                "class ClassSingleBase",
                "class ClassMultipleBase",
                "class ClassWithMeta",
                "class ClassWithMethods",
                "class ClassForTypedTests",
            ],
        ),
        (
            ["--async"],
            [
                "async def async_func",
                "async def async_func_fully_typed",
            ],
        ),
        # This doesn't make sense, so should return []
        (
            ["--async", "--class"],
            [],
        ),
        # Various typing options
        (
            ["--typed"],
            [
                "def func_type_annotations",
                "def function_with_non_pep_0484_annotation",
                "def complex_annotations",
                "def func_fully_typed",
                "async def async_func_fully_typed",
                "def func_partially_typed",
                "def func_partially_typed_no_typed_return",
                "def func_partially_typed_only_typed_return",
                "def func_typed_no_params",
            ],
        ),
        (
            ["--typed", "--async"],
            [
                "async def async_func_fully_typed",
            ],
        ),
        (
            ["--untyped"],
            [
                "def func_no_args",
                "def func_positional_args",
                "async def async_func",
                "def func_default_args",
                "def func_arbitrary_positional_args",
                "def func_arbitrary_keyword_args",
                "def func_arbitrary_args",
                "def func_positional_only_args",
                "def func_keyword_only_args",
            ],
        ),
        (
            ["--partially-typed"],
            [
                "def func_partially_typed",
                "def func_partially_typed_no_typed_return",
                "def func_partially_typed_only_typed_return",
            ],
        ),
        (
            ["--fully-typed"],
            [
                "def func_type_annotations",
                "def function_with_non_pep_0484_annotation",
                "def complex_annotations",
                "def func_fully_typed",
                "async def async_func_fully_typed",
                "def func_typed_no_params",
            ],
        ),
        # Test against methods
        (
            ["--typed", "*.*"],
            [
                "def method_types",
                "def __init__",
                "def method_fully_typed",
                "def method_partially_typed",
            ],
        ),
        (
            ["--untyped", "*.*"],
            [
                "def __init__",
                "def method_positional_only_args",
                "def method_keyword_only_args",
                "async def async_method",
                "def method_untyped",
            ],
        ),
        (
            ["--fully-typed", "*.*"],
            ["def method_types", "def __init__", "def method_fully_typed"],
        ),
        (
            ["--partially-typed", "*.*"],
            ["def method_partially_typed"],
        ),
        # Documented and undocumented
        (
            ["--documented"],
            [
                "def func_no_args",
                "def func_positional_args",
            ],
        ),
        (
            ["--undocumented", "func_arbitrary_*"],
            [
                "def func_arbitrary_positional_args",
                "def func_arbitrary_keyword_args",
                "def func_arbitrary_args",
            ],
        ),
        (
            ["--documented", "*.*"],
            [
                "def method_fully_typed",
                "def method_partially_typed",
            ],
        ),
        (
            ["--undocumented", "*.method_*"],
            [
                "def method_types",
                "def method_positional_only_args",
                "def method_keyword_only_args",
                "def method_untyped",
            ],
        ),
    ),
)
def test_filters(args, expected):
    runner = CliRunner()
    full_args = args + [
        "-s",
        "-f",
        str(pathlib.Path(__file__).parent / "example_symbols.py"),
    ]
    result = runner.invoke(
        cli,
        full_args,
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    # Remove # File: lines and blank lines
    lines = [
        line.strip()
        for line in result.stdout.splitlines()
        if line.strip() and not line.startswith("# File:")
    ]
    # We only match up to the opening "("
    defs = [line.split("(")[0] for line in lines]
    assert defs == expected

    # Test the --count option too
    expected_count = len(expected)
    result2 = runner.invoke(
        cli,
        full_args + ["--count"],
        catch_exceptions=False,
    )
    assert result2.exit_code == 0
    assert result2.stdout.strip() == str(expected_count)



import ast
import click
import pathlib

from .lib import code_for_node, find_symbol_nodes, read_file, type_summary


@click.command()
@click.version_option()
@click.argument("symbols", nargs=-1)
@click.option(
    "files",
    "-f",
    "--file",
    type=click.Path(file_okay=True, dir_okay=False),
    multiple=True,
    help="Files to search",
)
@click.option(
    "directories",
    "-d",
    "--directory",
    type=click.Path(file_okay=False, dir_okay=True, resolve_path=True),
    multiple=True,
    help="Directories to search",
)
@click.option(
    "-s",
    "--signatures",
    is_flag=True,
    help="Show just function and class signatures",
)
@click.option(
    "--docstrings",
    is_flag=True,
    help="Show function and class signatures plus docstrings",
)
@click.option(
    "--count",
    is_flag=True,
    help="Show count of matching symbols",
)
@click.option(
    "--silent",
    is_flag=True,
    help="Silently ignore Python files with parse errors",
)
@click.option(
    "async_",
    "--async",
    is_flag=True,
    help="Filter async functions",
)
@click.option(
    "--function",
    is_flag=True,
    help="Filter functions",
)
@click.option(
    "class_",
    "--class",
    is_flag=True,
    help="Filter classes",
)
@click.option(
    "--documented",
    is_flag=True,
    help="Filter functions with docstrings",
)
@click.option(
    "--undocumented",
    is_flag=True,
    help="Filter functions without docstrings",
)
@click.option(
    "--typed",
    is_flag=True,
    help="Filter functions with type annotations",
)
@click.option(
    "--untyped",
    is_flag=True,
    help="Filter functions without type annotations",
)
@click.option(
    "--partially-typed",
    is_flag=True,
    help="Filter functions with partial type annotations",
)
@click.option(
    "--fully-typed",
    is_flag=True,
    help="Filter functions with full type annotations",
)
def cli(
    symbols,
    files,
    directories,
    signatures,
    docstrings,
    count,
    silent,
    async_,
    function,
    class_,
    documented,
    undocumented,
    typed,
    untyped,
    partially_typed,
    fully_typed,
):
    """
    Find symbols in Python code and print the code for them.

    Example usage:

    \b
        # Search current directory and subdirectories
        symbex my_function MyClass

    \b
        # Search using a wildcard
        symbex 'test_*'

    \b
        # Find a specific class method
        symbex 'MyClass.my_method'

    \b
        # Find class methods using wildcards
        symbex '*View.handle_*'

    \b
        # Search a specific file
        symbex MyClass -f my_file.py

    \b
        # Search within a specific directory and its subdirectories
        symbex Database -d ~/projects/datasette

    \b
        # View signatures for all symbols in current directory and subdirectories
        symbex -s

    \b
        # View signatures for all test functions
        symbex 'test_*' -s

    \b
        # View signatures for all async functions with type definitions
        symbex --async --typed -s

    \b
        # Count the number of --async functions in the project
        symbex --async --count
    """
    if count or docstrings:
        signatures = True
    # Show --help if no filter options are provided:
    if not any(
        [
            symbols,
            signatures,
            async_,
            function,
            class_,
            documented,
            undocumented,
            typed,
            untyped,
            partially_typed,
            fully_typed,
        ]
    ):
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        ctx.exit()
    # Default to '*' if --signatures or filters are provided without symbols
    if (
        any(
            [
                signatures,
                async_,
                function,
                class_,
                documented,
                undocumented,
                typed,
                untyped,
                partially_typed,
                fully_typed,
            ]
        )
        and not symbols
    ):
        symbols = ["*"]
    if not files and not directories:
        directories = ["."]

    def iterate_files():
        yield from (pathlib.Path(f) for f in files)
        for directory in directories:
            for path in pathlib.Path(directory).rglob("*.py"):
                if path.is_file():
                    yield path

    # Filter symbols by type
    def filter(node: ast.AST) -> bool:
        return True

    # If any --filters were supplied, handle them:
    if any(
        [
            async_,
            function,
            class_,
            documented,
            undocumented,
            typed,
            untyped,
            partially_typed,
            fully_typed,
        ]
    ):

        def filter(node: ast.AST) -> bool:
            # Filters must ALL match
            if async_ and not isinstance(node, ast.AsyncFunctionDef):
                return False
            if function and not isinstance(
                node, (ast.FunctionDef, ast.AsyncFunctionDef)
            ):
                return False
            if class_ and not isinstance(node, ast.ClassDef):
                return False
            if documented and not ast.get_docstring(node):
                return False
            if undocumented and ast.get_docstring(node):
                return False
            summary = type_summary(node)
            # if no summary, type filters all fail
            if not summary and (typed or untyped or partially_typed or fully_typed):
                return False
            # Apply type filters
            if typed and not summary.partially:
                return False
            if untyped and summary.partially:
                return False
            if partially_typed and not (summary.partially and not summary.fully):
                return False
            if fully_typed and not summary.fully:
                return False

            return True

    pwd = pathlib.Path(".").resolve()
    num_matches = 0
    for file in iterate_files():
        code = read_file(file)
        try:
            nodes = find_symbol_nodes(code, str(file), symbols)
        except SyntaxError as ex:
            if not silent:
                click.secho(f"# Syntax error in {file}: {ex}", err=True, fg="yellow")
            continue
        for node, class_name in nodes:
            if not filter(node):
                continue
            if count:
                num_matches += 1
                continue
            # If file is within pwd, print relative path
            if pwd in file.resolve().parents:
                path = file.resolve().relative_to(pwd)
            else:
                # else print absolute path
                path = file.resolve()
            snippet, line_no = code_for_node(
                code, node, class_name, signatures, docstrings
            )
            bits = ["# File:", path]
            if class_name:
                bits.extend(["Class:", class_name])
            bits.extend(["Line:", line_no])
            print(*bits)
            print(snippet)
            print()
    if count:
        click.echo(num_matches)



# Tests for "symbex -s", using content of example_code.py
import pathlib
import pytest
import re
from click.testing import CliRunner

from symbex.cli import cli


@pytest.fixture
def symbols_text():
    runner = CliRunner()
    args = ["-s", "-f", str(pathlib.Path(__file__).parent / "example_symbols.py")]
    result = runner.invoke(cli, args, catch_exceptions=False)
    assert result.exit_code == 0
    return result.stdout


@pytest.mark.parametrize(
    "name,expected",
    (
        ("func_no_args", "def func_no_args()"),
        ("func_positional_args", "def func_positional_args(a, b, c)"),
        ("async_func", "async def async_func(a, b, c)"),
        ("func_default_args", "def func_default_args(a, b=2, c=3)"),
        ("func_arbitrary_positional_args", "def func_arbitrary_positional_args(*args)"),
        ("func_arbitrary_keyword_args", "def func_arbitrary_keyword_args(**kwargs)"),
        ("func_arbitrary_args", "def func_arbitrary_args(*args, **kwargs)"),
        ("func_positional_only_args", "def func_positional_only_args(a, /, b, c)"),
        ("func_keyword_only_args", "def func_keyword_only_args(a, *, b, c)"),
        ("func_type_annotations", "def func_type_annotations(a: int, b: str) -> bool"),
        ("ClassNoBase", "class ClassNoBase"),
        ("ClassSingleBase", "class ClassSingleBase(int)"),
        ("ClassMultipleBase", "class ClassMultipleBase(int, str)"),
        ("ClassWithMeta", "class ClassWithMeta(metaclass=type)"),
        (
            "function_with_non_pep_0484_annotation",
            "def function_with_non_pep_0484_annotation(x: ?, xx: ?, yy: ?, y: ?, zz: float) -> ?",
        ),
        (
            "complex_annotations",
            "complex_annotations(code: str, symbols: Iterable[str]) -> List[Tuple[(AST, Optional[str])]]",
        ),
    ),
)
def test_symbols(name, expected, symbols_text):
    # For error reporting try and find the relevant bit
    likely_line = [
        line
        for line in symbols_text.split("\n")
        if (f"{name}(" in line or line.startswith(f"class {name}"))
    ][0]
    assert expected in symbols_text, "\nexpected:\t{}\ngot:\t\t{}".format(
        expected, likely_line
    )
    # Special case to ensure we don't get ClassNoBase()
    assert "ClassNoBase()" not in symbols_text


def test_method_symbols():
    runner = CliRunner()
    args = [
        "*.async*",
        "-s",
        "-f",
        str(pathlib.Path(__file__).parent / "example_symbols.py"),
    ]
    result = runner.invoke(cli, args, catch_exceptions=False)
    assert result.exit_code == 0
    assert result.stdout == (
        "# File: tests/example_symbols.py Class: ClassWithMethods Line: 91\n"
        "    async def async_method(a, b, c)\n"
        "\n"
    )


def test_docstrings():
    runner = CliRunner()
    args = [
        "*.*",
        "*",
        "--documented",
        "--docstrings",
        "-f",
        str(pathlib.Path(__file__).parent / "example_symbols.py"),
    ]
    result = runner.invoke(cli, args, catch_exceptions=False)
    assert result.exit_code == 0
    expected = """
# File: tests/example_symbols.py Line: X
def func_no_args()
    "This has a single line docstring"

# File: tests/example_symbols.py Line: X
def func_positional_args(a, b, c)
    \"\"\"This has a
    multi-line docstring\"\"\"

# File: tests/example_symbols.py Class: ClassForTypedTests Line: X
    def method_fully_typed(self, a: int, b: str) -> bool
        "Single line"

# File: tests/example_symbols.py Class: ClassForTypedTests Line: X
    def method_partially_typed(self, a: int, b) -> bool
        \"\"\"Multiple
        lines\"\"\"
    """.strip()
    actual = result.stdout.strip()
    # Replace 'Line \d' with 'Line X' before comparison using re
    actual = re.sub(r"Line: \d+", "Line: X", actual)
    assert actual == expected
